#pragma once

#include <coroutine>
#include <iostream>
#include <memory>
#include <optional>
#include <stdexcept>
#include <utility>
#include <vector>

#include <Game/Core.hpp>

#include <bt/behavior_node.hpp>
#include <bt/interact.hpp>
#include <bt/inventory.hpp>
#include <bt/login.hpp>
#include <bt/magic.hpp>
#include <bt/mainscreen.hpp>
#include <bt/navigation.hpp>
#include <bt/node_task.hpp>
#include <bt/player.hpp>
#include <bt/script.hpp>

namespace bt {
template <typename Context> class tree_builder {
public:
  tree_builder() noexcept : _root(nullptr) {}

  tree_builder &begin_sequence() {
    _node_stack.push_back(std::make_unique<sequence_node<Context>>());
    _children.emplace_back();
    return *this;
  }

  tree_builder &end_sequence() {
    if (_node_stack.empty()) {
      throw std::runtime_error("Mismatched begin_sequence/end_sequence");
    }

    auto node = std::move(_node_stack.back());
    _node_stack.pop_back();
    auto children = std::move(_children.back());
    _children.pop_back();

    for (auto &&child : children) {
      static_cast<composite_node<Context> *>(node.get())
          ->add_child(std::move(child));
    }

    if (_node_stack.empty()) {
      _root = std::move(node);
    } else {
      _children.back().push_back(std::move(node));
    }

    return *this;
  }

  tree_builder &begin_selector() {
    _node_stack.push_back(std::make_unique<selector_node<Context>>());
    _children.emplace_back();
    return *this;
  }

  tree_builder &end_selector() {
    if (_node_stack.empty()) {
      throw std::runtime_error("Mismatched begin_selector/end_selector");
    }

    auto node = std::move(_node_stack.back());
    _node_stack.pop_back();
    auto children = std::move(_children.back());
    _children.pop_back();

    for (auto &&child : children) {
      static_cast<composite_node<Context> *>(node.get())
          ->add_child(std::move(child));
    }

    if (_node_stack.empty()) {
      _root = std::move(node);
    } else {
      _children.back().push_back(std::move(node));
    }

    return *this;
  }

  template <template<typename...> class Node, typename... Args>
  tree_builder &leaf(Args&&... args) {
    if (_node_stack.empty()) {
      throw std::runtime_error("Leaf must be within a sequence or selector");
    }

    _children.back().push_back(std::make_unique<Node<Context, Args...>>(std::forward<Args>(args)...));
    return *this;
  }

  template <typename TickFn> tree_builder &action(TickFn tick) {
    if (_node_stack.empty()) {
      throw std::runtime_error("Action must be within a sequence or selector");
    }

    _children.back().push_back(
        std::make_unique<action_node<Context, TickFn>>(tick));
    return *this;
  }

  template <typename ConditionFn>
  tree_builder &conditional(ConditionFn condition) {
    if (_node_stack.empty()) {
      throw std::runtime_error(
          "Condition must be within a sequence or selector");
    }

    _children.back().push_back(
        std::make_unique<conditional_node<Context, ConditionFn>>(condition));
    return *this;
  }

  tree_builder &invert() {
    if (_node_stack.empty()) {
      throw std::runtime_error("Invert must be within a sequence or selector");
    }

    if (_children.back().empty()) {
      throw std::runtime_error("Invert requires a preceding node");
    }

    auto child = std::move(_children.back().back());
    _children.back().pop_back();
    _children.back().push_back(
        std::make_unique<inverter_node<Context>>(std::move(child)));

    return *this;
  }

  std::unique_ptr<behavior_node<Context>> build() {
    if (!_node_stack.empty()) {
      throw std::runtime_error("Mismatched begin/end");
    }

    return std::move(_root);
  }
  
private:
  std::unique_ptr<behavior_node<Context>> _root;
  std::vector<std::unique_ptr<behavior_node<Context>>> _node_stack;
  std::vector<std::vector<std::unique_ptr<behavior_node<Context>>>> _children;
};
} // namespace bt